/*********************************************************************
 *
 * Authors: Vincenzo Ciaschini - Vincenzo.Ciaschini@cnaf.infn.it 
 *
 * Copyright (c) 2002, 2003 INFN-CNAF on behalf of the EU DataGrid.
 * For license conditions see LICENSE file or
 * http://www.edg.org/license.html
 *
 * Parts of this code may be based upon or even include verbatim pieces,
 * originally written by other people, in which case the original header
 * follows.
 *
 *********************************************************************/
/***************************************************************************
 *  filename  : GSISocketServer.cpp
 *  authors   : Salvatore Monforte <salvatore.monforte@ct.infn.it>
 *  copyright : (C) 2001 by INFN
 ***************************************************************************/

// $Id:

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/** The globus secure shell API definitions. */
#include "gssapi.h"
#include <memory.h>
#include <time.h>
#include <stdio.h>

#include <openssl/evp.h>
#include "openssl/buffer.h"
#include "openssl/crypto.h"
#include "openssl/objects.h"

/** Functionalities for transmitting and receiveing tokens. */
#include "tokens.h"
#include "newca.h"
/** This class header file. */
#include "Server.h"

#include <sys/types.h>
#include <unistd.h>

static char *
get_globusid(gss_cred_id_t handle)
{
  char *globusid;
  char *globusid_tmp;
  gss_name_t server_name = GSS_C_NO_NAME;
  gss_buffer_desc server_buffer_desc = GSS_C_EMPTY_BUFFER;
  gss_buffer_t server_buffer = &server_buffer_desc;
  OM_uint32 major_status = 0;
  OM_uint32 minor_status = 0;
  OM_uint32 minor_status2 = 0;

  if ((major_status = gss_inquire_cred(&minor_status,
				       handle,
				       &server_name,
				       NULL, NULL, NULL)) == GSS_S_COMPLETE) {
    major_status = gss_display_name(&minor_status,
				    server_name, server_buffer, NULL);
    gss_release_name(&minor_status2, &server_name);
  }
  /*
   * The gssapi_cleartext does not implement gss_inquire_cred,
   * so fall back to using environment variable.
   */
  if (major_status == GSS_S_COMPLETE) {
    globusid = (char *)server_buffer_desc.value;
  }
  else {
    return NULL;
  }
  globusid_tmp = strdup(globusid);
  
  if (server_buffer_desc.value) {
    gss_release_buffer(&minor_status2, server_buffer);
  }
  return globusid_tmp;
}


/**
 * Constructor.
 * @param p the secure server port.
 * @param b the backlog, that is the maximum number of outstanding connection requests.
 */
GSISocketServer::GSISocketServer(int p, int v, int b, bool m) :
  gsi_logfile(NULL), version(v), own_subject(""), own_ca(""),
  peer_subject(""), peer_ca(""), own_key(NULL), peer_key(NULL),
  port(p), opened(false), credential(GSS_C_NO_CREDENTIAL),
  context(GSS_C_NO_CONTEXT), backlog(b), newopened(false), mustclose(m),
  conflags(0)
{
  OBJ_create("0.9.2342.19200300.100.1.1","USERID","userId");
}

bool 
GSISocketServer::Open()
{
  OM_uint32      major_status, minor_status;
  bool result = false;
  struct sockaddr_in myaddr_in;

  memset ((char *)&myaddr_in, 0, sizeof(struct sockaddr_in));
  
  myaddr_in.sin_family = AF_INET;
  myaddr_in.sin_addr.s_addr = INADDR_ANY;
  myaddr_in.sin_port=htons(port);

  if((sck = socket (AF_INET, SOCK_STREAM, 0)) == -1)
    return false;

  opened = true;
  unsigned int value = 1;
	
  setsockopt( sck, SOL_SOCKET, SO_REUSEADDR, (void *) &value, sizeof(int) );
  result = ((bind(sck, (struct sockaddr*)&myaddr_in, sizeof(struct sockaddr_in)) != -1) && (listen(sck, backlog) != -1));

  if (result) {
    major_status = globus_gss_assist_acquire_cred(&minor_status,
						  GSS_C_BOTH,
						  &credential);
    if(!GSS_ERROR(major_status)) {
      char *tmp;

      tmp = get_globusid(credential);
      if (tmp)
	own_subject = std::string(tmp);

      tmp = getMCA(credential, version);

      if (tmp)
	own_ca = std::string(tmp);

      own_key = get_private_key(credential, version);

      return true;
    }

    if (gsi_logfile)
      globus_gss_assist_display_status(gsi_logfile,
				       "Failed to acquire credentials:",
				       major_status, minor_status, 0);
  }
  return false;
}

/**
 * Destructor.
 */
GSISocketServer::~GSISocketServer()
{
  Close();
}


/**
 * Close the connection.
 */
void 
GSISocketServer::Close()
{
  OM_uint32 status;

  if (context != GSS_C_NO_CONTEXT)
    gss_delete_sec_context(&status, &context, GSS_C_NO_BUFFER);
  context = GSS_C_NO_CONTEXT;
  if (credential != GSS_C_NO_CREDENTIAL)
    gss_release_cred(&status, &credential);
  credential = GSS_C_NO_CREDENTIAL;
  if (newopened)
    close(newsck);
  newopened=false;
  if (opened)
    close(sck);
  opened=false;
}

void GSISocketServer::CloseListener(void)
{
  if (opened)
    close(sck);
  opened = false;
}

void GSISocketServer::CloseListened(void)
{
  if (newopened)
    close(newsck);
  newopened = false;
}

void 
GSISocketServer::SetFlags(OM_uint32 flags)
{
  conflags = flags;
}

/**
 * Accept the GSI Authentication.
 * @param sock the socket for communication.
 * @param ctx the authorization context.
 * @return the context identifier. 
 */
bool
GSISocketServer::AcceptGSIAuthentication()
{
  OM_uint32      major_status = 0, minor_status = 0, status=0;
  OM_uint32      ret_flags = 0;
  int            token_status = 0;
  char           *name = NULL;
  bool           return_status = false;
  gss_cred_id_t  delegated_cred = GSS_C_NO_CREDENTIAL;
  char           *tmp;

  if (!newopened)
    return false;

  if (context != GSS_C_NO_CONTEXT)
    gss_delete_sec_context(&status, &context, GSS_C_NO_BUFFER);
  context = GSS_C_NO_CONTEXT;

  major_status =
    globus_gss_assist_accept_sec_context(&minor_status, &context,
					 credential, &name,
					 &ret_flags, NULL,
					 &token_status, &delegated_cred,
					 &get_token, (void *) &newsck,
					 &send_token, (void *) &newsck);

  if (GSS_ERROR(major_status)) {
    if (gsi_logfile)
      fprintf(gsi_logfile,"Major: %x, minor: %x\n", major_status, minor_status); 
      globus_gss_assist_display_status(gsi_logfile,
				       "Failed to establish security context (accept):",
				       major_status, minor_status, token_status);
    return false;
  }

   if ((ret_flags & conflags) != conflags) {
     if (gsi_logfile)
       fprintf(gsi_logfile, "Flags Mismatch:\nExpected: %d\nReceived:%d\n",
	       conflags, (ret_flags & conflags));
     return false;
   }

  peer_subject = name;

  tmp = get_peer_CA(context, version);

  if (tmp)
    peer_ca = std::string(tmp);

  peer_key = get_delegated_public_key(context, version);

  return true;
}

/**
 * Listen for incoming connection requests.
 * Accept incoming requests and redirect communication on a dedicated port.
 * @param a a reference to the secure GSI Socket Agent sent by Client.
 * @return the GSI Socket Agent redirecting communication on a dedicated port.
 */
bool 
GSISocketServer::Listen()
{
  struct sockaddr_in peeraddr_in;
  socklen_t addrlen = sizeof(struct sockaddr_in);

  if (!opened) {
    fprintf(gsi_logfile,"Socket unopened!\n");
    return false;
  }

  newsck = accept(sck, (struct sockaddr*)&peeraddr_in, &addrlen);

  if (newsck != -1) {
    newopened = true;
    return AcceptGSIAuthentication();
  }
  else
    return false;
}

/**
 * Send a string value.
 * @param s the string value to send.
 * @return true on success, false otherwise.
 */ 
bool 
GSISocketServer::Send(const std::string s)
{
  bool return_status;

#if 0
  if (mustclose && opened) {
    close(sck);
    opened = false;
  }
#endif

  if (return_status = !(context == GSS_C_NO_CONTEXT)) {

    OM_uint32        maj_stat, min_stat;
     
    int token_status;
    maj_stat = globus_gss_assist_wrap_send (&min_stat, context,
					    const_cast<char *>(s.c_str()),
					    s.length(), &token_status,
					    send_token, &newsck, gsi_logfile);

    return_status = !GSS_ERROR(maj_stat);
  }

  return return_status;
}


/**
 * Receive a string value.
 * @param s the string to fill.
 * @return true on success, false otherwise.
 */
bool 
GSISocketServer::Receive(std::string& s)
{
  bool return_status;

  OM_uint32 maj_stat, min_stat;

  char  *message;
  size_t length;
  int    token_status;

#if 0
  if (mustclose && opened) {
    close(sck);
    opened = false;
  }
#endif

  maj_stat = globus_gss_assist_get_unwrap(&min_stat, context,
					  &message, &length, &token_status, 
					  get_token, &newsck, gsi_logfile);

  if (return_status = !GSS_ERROR(maj_stat)) 
    s = std::string(message,length);
  else
    globus_gss_assist_display_status(gsi_logfile, 
				     "GSS authentication failure ",
				     maj_stat, min_stat, token_status); 
  return return_status;
}
