/*********************************************************************
 *
 * Authors: Vincenzo Ciaschini - Vincenzo.Ciaschini@cnaf.infn.it 
 *
 * Copyright (c) 2002, 2003 INFN-CNAF on behalf of the EU DataGrid.
 * For license conditions see LICENSE file or
 * http://www.edg.org/license.html
 *
 * Parts of this code may be based upon or even include verbatim pieces,
 * originally written by other people, in which case the original header
 * follows.
 *
 *********************************************************************/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <netdb.h>
#include <strings.h>
#include <netinet/in.h>
#include <sys/types.h>
#include <sys/socket.h>
#include "globus_config.h"
#include "gssapi.h"
#include "gssapi_compat.h"
#include "globus_gss_assist.h"

#include <stdio.h>
#include <time.h>
#include <openssl/x509.h>
#include <openssl/asn1.h>

#include "format.h"
#include "newca.h"

#include "Client.h"

#include <iostream>

bool net_connect(const std::string &hostname, int port,
		 const std::string &contact, const std::string &vo,
		 const std::string &command, std::string &buffer,
		 int version, int num, std::string &errmsg, bool quiet);
static bool verify_data(const std::string &vo, const std::string &host,
			const std::string &user, const std::string &userca,
			const std::string &serverca,
			std::string &message, const EVP_PKEY *key,
			std::string &errmsg, int num, bool quiet);

/*
 * Function:
 *   verify_data(vo, host, user, userca, serverca, message, length, key)
 *
 * Description:
 *   This function takes a message returned by the VOMS server and verifies
 *   the correctness of the data. This means that it verifies the signature,
 *   the subjects and CAs of both client and server, that the information is
 *   neither stale or premature, etc...
 *
 * Parameters:
 *   'vo'       - The VO to whom the VOMS server belongs.
 *   'host'     - The subject of the server's certificate.
 *   'user'     - The subject of the user's certificate.
 *   'userca'   - The subject of the CA that issued the user's certificate.
 *   'serverca' - The subject of the CA that issued the server's certificate.
 *   'message'  - The message to verify.
 *   'length'   - The length of the message.
 *   'key'      - The public key of the server.
 *
 * Result:
 *   Failure:
 *     0.
 *   Success:
 *     1.
 *
 * NOTE:
 *   This function takes for granted that the identity of the host has already
 *   been verified, and that it can rely on the correctness of the public key.
 */
static bool
verify_data(const std::string &vo, const std::string &host, const std::string &user,
	    const std::string &userca, const std::string &serverca,
	    std::string &message, const EVP_PKEY *key, std::string &errmsg, int num,
	    bool quiet)
{
  struct collection coll;
  bool result = false;
  std::string error;

  if (host.empty() || user.empty() || userca.empty() ||
      serverca.empty() || message.empty() || !key || vo.empty())
    return false;

  while (message[0] == '2') {
    if (num)
      errmsg += "WARNING!\n";
    else
      errmsg += "ERROR!\n";
    std::string::size_type t = message.find('\1');
    if (t != std::string::npos) {
      errmsg += message.substr(1, t-1);
      message = message.substr(t+1);
    }
    return false;
  }

  while (message[0] == '1') {
    errmsg += "WARNING!\n";
    std::string::size_type t = message.find('\1');
    if (t != std::string::npos) {
      errmsg += message.substr(1, t-1);
      message = message.substr(t+1);
    }
  }

  if (!num) {
    if (message[0] == '0') {
      errmsg += "WARNING!\n";
      std::string::size_type t = message.find('\1');
      if (t != std::string::npos) {
	errmsg += message.substr(1, t-1);
	message = message.substr(t+1);
      }
    }
  }

  /* Parse the date. This also verifies the signature. */
  if (!unformat(message, key, coll)) {
    error = "Message uncorrectly formatted!\n";
    return false;
  }

  unsigned char *timebuf;
  ASN1_TIME *time1 = NULL, *time2 = NULL;

  timebuf = (unsigned char *)coll.date1.c_str();
  time1 = d2i_ASN1_TIME(NULL, &timebuf, coll.date1.size());

  timebuf = (unsigned char *)coll.date2.c_str();
  time2 = d2i_ASN1_TIME(NULL, &timebuf, coll.date2.size());

  if (!time1 || !time2) {
    error = "Can't interpret date!\n";
    goto error;
  }

  /* Verifies the date constraints */
  time_t ctime;
  time (&ctime);
  ctime += 300;

  if (((X509_cmp_current_time(time1) >=0) &&
       (X509_cmp_time(time1, &ctime) >=  0)) || 
      (X509_cmp_current_time(time2) <= 0)) {
    error = "Date mismatch problem!\n";
    goto error;
  }


  /* Verifies the subjects and CAs and VO */
  if (coll.voname == vo) {
    if (coll.server == host) {
      if (coll.serverca == serverca) {
	if (coll.user == user) {
	  if (coll.userca == userca)
	    result = true;
	  else 
	    error = "User CA mismatch!\n" +  coll.userca + "\nIn place of:\n" +
	      userca + ".\n";
	}
	else
	  error = "User subject mismatch!\n" + coll.user + "\nIn place of:\n" +
	    user + "\n";
      }
      else
	error = "Server CA mismatch!\n" + coll.serverca + "\nIn place of:\n" +
	  serverca + "\n";
    }
    else
      error = "Server subject mismatch!\n" +  coll.server + "\nIn place of:\n" +
	host + "\n";
  }
  else
    error = "VO name mismatch!\n" + coll.voname + "\nIn place of:\n" + vo + "\n";



 error:
  if (!quiet)
    std::cerr << error << std::flush;
  if (time1)
    ASN1_TIME_free(time1);
  if (time2)
    ASN1_TIME_free(time2);
  return result;
}

/*
 * Function:
 *   net_connect(hostname, port, contact, vo, command, commandlen, buffer, buflen)
 *
 * Description:
 *   This function connects to the VOMS server, submits a command, gets the
 *   results, verifies them and finally returns them.
 *
 * Parameters:
 *   'hostname'   - This is the hostname of the server to contact.
 *   'port'       - This is the port on which the server is listening.
 *   'contact'    - This is the subject of the server's certificate.
 *   'vo'         - The VO to whom the VOMS server belongs.
 *   'command'    - This is the command to be executed.
 *   'commandlen' - This is the command's length
 *   'buffer'     - OUTPUT PARAMETER: This buffer contains the result of
 *                  the command sent to the server.
 *   'buflen'     - OUTPUT PARAMETER: This is the length of the 'buffer'
 *                  parameter.
 *
 * Result:
 *   Failure:
 *     0.
 *   Success:
 *     Non zero. Furthermore, the output parameters are properly initialized.
 *
 * NOTE: The memory pointed by the 'buffer' parameter may very well be
 * overwritten by consecutive calls to this function.
 */
bool
net_connect(const std::string &hostname, int port, const std::string &contact,
	    const std::string &vo, const std::string &command, std::string &buffer,
	    int version, int num, std::string &errmsg, bool quiet)
{
#ifdef DEBUG
  cout << hostname << ":" << port << ":" << version << endl << flush;
#endif
  GSISocketClient sock(hostname, port, version);

  sock.RedirectGSIOutput(stderr);
  sock.ServerContact(contact);
  sock.SetFlags(GSS_C_MUTUAL_FLAG | GSS_C_CONF_FLAG | GSS_C_INTEG_FLAG);

  if(!sock.Open()) {
    if (!quiet)
      std::cerr << "Unable to connect to server" << std::endl << std::flush;
    return false;
  }

#ifdef DEBUG
  if (!quiet) {
    cout << "***********" << endl
	 << "OWN  SUBJ:" << sock.own_subject << endl
	 << "OWN  CA  :" << sock.own_ca << endl
	 << "PEER SUBJ:" << sock.peer_subject << endl
	 << "PEER CA  :" << sock.peer_ca << endl << flush;
  }
#endif          

  std::string username = sock.own_subject;
  std::string userca   = sock.own_ca;
  std::string serverca = sock.peer_ca;
  std::string server   = sock.peer_subject;

#ifdef DEBUG
  if (!quiet)
    cout << username << endl << userca << endl << flush;
#endif

  if (username.empty()) {
    if (!quiet)
      std::cerr << "Unable to identify self!" << std::endl << std::flush;
    return false;
  }

  sock.Send(command);
  sock.Receive(buffer);

  if (!quiet) {
    std::cout << server << std::endl << std::flush;
    std::cout << serverca << std::endl << std::flush;
  }
  /* Verify results */

  bool r = verify_data(vo, contact, username, userca, serverca, buffer, sock.peer_key, errmsg, num, quiet);
  sock.Close();
  return r;
}   
